
#ifndef _MAIN_GAME_H
#define _MAIN_GAME_H


#include "partypack/vector2d.h"
#include <vector>
#include <list>
#include "GameState.h"

using namespace std;

class Level;
class alBitmap;
class SpriteAnimator;


class MainGame : public GameState {
	public:
		MainGame();
		~MainGame();

		void setup(Level *level, bool machinegun);
		void update();
		void draw(alBitmap &bitmap);

		int getScore() { return score; }
		bool paused;

	private:
		void reload();
		void offset(vector2f pos, bool inscope, vector2f &v, bool &b);

		class Pin {
			public:
				Pin(vector2f loc, MainGame *game);
				~Pin();
				void draw(alBitmap &bitmap, vector2f offset, bool inscope);
			private:
				BITMAP *pin, *pinshadow, *pinsmall, *pinsmallshadow;
				vector2f pos;
				MainGame *game;
		};
		class Troll {
			public:
				Troll(int x, int y, MainGame *game);
				~Troll();
				void update();
				void draw(alBitmap &bitmap, bool inscope);
				bool shoot(vector2f p);
			private:
				SpriteAnimator *anim, *animsmall, *animhit;
				vector2f pos,delta;
				MainGame *game;
				int frame;
				int health;
				bool alive;
				float r;
				vector<Pin> hits;
		};
		class Fire {
			public:
				Fire(float x, float y, float dx, MainGame *game);
				~Fire();
				bool update();
				void draw(alBitmap &bitmap, bool inscope);
				friend class MainGame;
			private:
				MainGame *game;
				BITMAP *flame, *flamesmall;
				vector2f pos,delta;
				int age;

		};
		class Explosion {
			public:
				Explosion(float x, float y, MainGame *game);
				~Explosion();
				bool update();
				void draw(alBitmap &bitmap, bool inscope);
				friend class MainGame;
			private:
				MainGame *game;
				SpriteAnimator *anim, *animsmall;
				vector2f pos,delta;
				int age;

		};
		friend class Pin;
		friend class Troll;
		friend class Fire;
		friend class Explosion;

		vector<Troll*> trolls;
		list<Fire*> flames;
		list<Explosion*> explosions;
		vector<Pin> missedpins;

		vector2f crosshair,crosshaird;
		vector2f scroll;
		alBitmap background;
		BITMAP *flame, *flamesmall, *smoke, *pin, *pin2;
		SAMPLE *gunshotfx, *reloadfx, *hitfx, *explodefx;

		int pins;
		int activepins;
		int lastshot;

		int shots, hits, misses;

		int trollsn, trollsleft;

		int score;
		bool machinegun,hasmachine;

		int time;

		bool canPlay;

};



#endif
